import numpy as np


def assert_dimension(data, required_dimension, name=None, msg=None):
    """
    Perform two checks:
    1. data must have .ndim attribute;
    2. data.ndim must return value of `required_dimension`;
    Otherwise raises ValueError.

    Name is subsititude in standard output message.
    if msg is not None, msg is returned with msg.format(name=name, required_dimension=required_dimension, got_dimension=data.ndim)
    """
    if data.ndim != required_dimension:
        if msg is None:
            msg = "{name} dimension is required to be {required_dimension}, however got {got_dimension}."
            raise ValueError(
                msg.format(
                    name=name,
                    required_dimension=required_dimension,
                    got_dimension=data.ndim))



from functools import wraps


def unbox(**kwargs):
  for name, type_ in kwargs.values():
    pass

  def warpper(func):
    def call(*args, **kwargs):
      func(*args, **kwargs)

    return call

  return warpper


import numpy as np


def fspecial(size, sigma):
    """Function to mimic the 'fspecial' gaussian MATLAB function.
    :returns result: 2d ndarray, gaussian filter. with shape `[1, size, size, 1]`.
    :type result: Numpy array
    """
    radius = size // 2
    offset = 0.0
    start, stop = -radius, radius + 1
    if size % 2 == 0:
        offset = 0.5
        stop -= 1
    x, y = np.mgrid[offset + start:stop, offset + start:stop]
    assert len(x) == size
    g = np.exp(-((x**2 + y**2) / (2.0 * sigma**2)))
    result = g / g.sum()
    result = result.reshape([result.shape[0], result.shape[1], 1, 1])
    return result



import click


@click.group()
def dxdata():
  pass


@dxdata.group()
def image():
  pass


@image.group()
def analysis():
  pass


@analysis.command()
@click.argument('target-path-file', type=click.Path(exists=True))
@click.argument('target-path-dataset', type=click.Path())
@click.option('--metric', '-m', type=str, multiple=True)
@click.option('--output', '-o', type=click.Path())
def single(target_path_file, target_path_dataset, metric, output):
  from ..statistics import analysis
  result = analysis({
      'type': 'h5',
      'path_file': target_path_file,
      'path_dataset': target_path_dataset
  }, None, metric, None, output)
  click.echo(str(result))



from enum import Enum
from functools import reduce
from typing import Iterable, Tuple

import numpy as np
import pandas as pd
from dxl.fs import Path


class Data:
    """
    Baseclass of data. Which is designed to be immutable, and processed by calling its methods to produce new data.
    """

    def __init__(self, data):
        self._data = data

    @classmethod
    def from_(cls, data):
        return cls(data)

    def to(self, constructor):
        return constructor(self.data())

    def data(self):
        return self._data

    @property
    def d(self):
        """
        Alias to self.data
        """
        return self._data


class DataIterable(Data):
    """
    Base class for iterable(list, tuple) of data.
    """

    def __init__(self, data: Iterable[Data]):
        super().__init__(data)

    def map(self, func) -> 'DataIterable':
        return DataIterable(list(map(func, self.d)))

    def filter(self, func) -> 'DataIterable':
        return DataIterable(list(filter(func, self.d)))

    def map_call(self, func_name, *args, **kwargs) -> 'DataIterable':
        """
        Call member method with func_name for each entry of data.
        """
        return self.map(lambda x: getattr(x, func_name)(*args, **kwargs))

    def map_to(self, constructor) -> 'DataIterable':
        return self.map(lambda x: constructor(x))

    def zip(self, data: 'DataIterable'):
        return DataIterable(zip(self.d, data.d))

    def first(self) -> Data:
        if isinstance(self.d, (list, tuple)):
            return self.d[0]
        else:
            return next(self.d)

    def flatten(self) -> 'DataIterable':
        result = []
        for d in self.data():
            if not isinstance(d, DataIterable):
                raise TypeError(
                    "Flatten requires Results of Results, got {}.".format(type(d)))
            result += d.to_list()
        return DataIterable(result)

    def merge(self, initializer=None) -> Data:
        """
        Call merge method on data, like reduce. 
        """
        result = initializer
        for o in self.data():
            if result is None:
                result = o
            else:
                result = result.merge(o)
        return result

    def to_list(self):
        return list(self.d)

    def __iter__(self):
        for d in self.data():
            yield d


class DataIterableWithKeys(DataIterable):
    def __init__(self, data: Iterable[Tuple]):
        if isinstance(data, DataIterable):
            super().__init__(data.data())
        elif isinstance(data, dict):
            super().__init__(data.items())
        else:
            super().__init__(data)

    def drop_keys(self) -> DataIterable:
        return self.map(lambda x: x[1])

    def to_dict(self):
        return {d[o]: d[1] for d in self.data()}

    def select(self, key) -> Data:
        for d in self.data():
            if d[0] == key:
                return d[1]
        raise KeyError("Key {} not found.".format(key))

    def select_all(self, key) -> DataIterable:
        result = []
        for d in self.data():
            if d[0] == key:
                result.append(d[1])
        return result


class DataNDArray(Data):
    @classmethod
    def convert_data_to_ndarray(cls, data):
        import numpy as np
        if isinstance(data, DataIterable):
            data = [d.data() for d in data.data()]
        if isinstance(data, Data):
            data = data.data()
        if isinstance(data, (list, tuple)):
            data = [d.data() if isinstance(d, Data) else d for d in data]
        return np.array(data)

    def __init__(self, data):
        super().__init__(self.convert_data_to_ndarray(data))

    @classmethod
    def from_(cls, data) -> 'DataNDArray':
        if isinstance(data, DataIterable):
            result = []
            for d in data:
                result.append(np.array(d.data()))
            return cls(np.array(result))


# class ParticleID(Enum):
#     Gamma = 22


# class ColumnNames:
#     Event = 'eventID'
#     Process = 'processName'
#     Particle = 'particalID'
#     SourceX = 'srcX'
#     SourceY = 'srcY'
#     SourceZ = 'srcZ'
#     X = 'x'
#     Y = 'y'
#     Z = 'z'
#     EnergyDeposit = 'energy'


# class ResultsDask(ResultBase):
#     def map(self, func) -> 'ResultsDask':
#         return Results(dask.delayed(map)(func, self.d))


# class ResultsNamedTuple(ResultBase):
#     pass


# class ResultsWithKeys(Results):
#     def __init__(self, data: Tuple[Tuple[str, ResultBase]]):
#         super().__init__(data)

#     def drop_keys(self) -> Results:
#         return self.map(lambda x: x[1])

#     def to_dict(self):
#         return {o[0]: o[1] for o in self.d}

#     def select(self, key):
#         for o in self.d:
#             if o[0] == key:
#                 return o[1]
#         raise KeyError("Key {} not found.".format(key))


# class ResultsWithUnknownKeys(ResultsDask):
#     def select(self, key):
#         return DataFrame(self.d.get_group(key))

#     def drop_keys(self) -> ResultsDask:
#         return self.map(lambda x: x[1])


# class Series(ResultBase):
#     def __init__(self, series: pd.Series):
#         super().__init__(series)

#     def columns(self, *cols):
#         return (self.d[c] for c in cols)

#     def position(self) -> 'Vec3':
#         return Vec3(*self.columns(ColumnNames.X,
#                                   ColumnNames.Y,
#                                   ColumnNames.Z))

#     def source_position(self) -> 'Vec3':
#         return Vec3(*self.columns(ColumnNames.SourceX,
#                                   ColumnNames.SourceY,
#                                   ColumnNames.SourceZ))

#     def energy_deposit(self) -> float:
#         return self.d[ColumnNames.EnergyDeposit]


# class DataFrame(ResultBase):
#     def __init__(self, dataframe: pd.DataFrame):
#         super().__init__(dataframe)

#     def first(self, *columns) -> Series:
#         return Series(self.d.iloc[0])

#     def split_by(self, column: str) -> ResultsWithKeys:
#         groups = self.d.groupby(column)
#         if isinstance(groups, DataFrameGroupBy):
#             return ResultsWithUnknownKeys(groups)
#         else:
#             return ResultsWithKeys(((k, DataFrame(groups.get_group(k).drop([column], axis=1))) for k in groups.groups))

#     def split_row(self) -> Results:
#         return Results((Series(self.d.iloc[i]) for i in range(self.d.shape[0])))

#     def merge(self, r) -> 'DataFrame':
#         if not isinstance(r, DataFrame):
#             raise TypeError(
#                 "Can not merge {} with {}.".format(__class__, type(r)))
#         return DataFrame(pd.concat([self.d, r.d], axis=0))

#     def to_event(self) -> 'Event':
#         if not ColumnNames.Event in self.d.columns:
#             return Event(self.d)


# class Vec3(ResultBase):
#     def __init__(self, x, y=None, z=None):
#         super().__init__(np.array([x, y, z]))

#     @property
#     def x(self) -> float:
#         return self.d[0]

#     @property
#     def y(self) -> float:
#         return self.d[1]

#     @property
#     def z(self) -> float:
#         return self.d[2]

#     def to_list(self):
#         return tuple(self.d)


# class EnergyDeposit(ResultBase):
#     def __init__(self, position, energy):
#         super().__init__((position, energy))

#     @property
#     def position(self):
#         return self.d[0]

#     @property
#     def energy(self):
#         return self.d[1]


# class Event(DataFrame):
#     def __init__(self, dataframe: pd.DataFrame):
#         super().__init__(dataframe)

#     def source_position(self) -> Vec3:
#         return self.first().source_position()

#     def first_position(self) -> Vec3:
#         return self.first().position()

#     def incident_direction(self) -> Vec3:
#         dp = self.first_position().d - self.source_position().d
#         return Vec3(*tuple(dp / np.linalg.norm(dp)))

#     def energy_deposit_list(self) -> Results:
#         return self.split_row().map(lambda s: EnergyDeposit(s.position(), s.energy_deposit()))



from pathlib import Path
from dxl.core.log import logger
import numpy as np


def load_npz(path):
  """
  Load npz file into dict and logging their shapes to info.
  """
  path = Path(path)
  logger.info('Loading npz file {}...'.format(str(path)))
  data = np.load(path)
  result = {}
  for k in data:
    result[k] = np.array(data[k])
    logger.info('Field: {}, shape: {}.'.format(k, result[k].shape))
  return result



from .core import Data, DataIterable, DataIterableWithKeys
from .geometry import VectorLowDim
import pandas as pd


class Series(Data):
  def __init__(self, series: pd.Series):
    super().__init__(series)

  def columns(self, *cols):
    return DataIterable([self.d[c] for c in cols])


class DataFrame(Data):
  def __init__(self, data):
    if not isinstance(data, pd.DataFrame):
      data = pd.DataFrame(data=data)
    super().__init__(data)

  def nb_rows(self):
    return self.data().shape[0]

  def first(self) -> Series:
    return Series(self.d.iloc[0])

  def split_by(self, column: str) -> DataIterableWithKeys:
    groups = self.d.groupby(column)
    return DataIterableWithKeys([(k,
                                  DataFrame((groups.get_group(k).drop(
                                      [column], axis=1))))
                                 for k in groups.groups])

  def split_row(self) -> DataIterable:
    return DataIterable(
        [Series(self.d.iloc[i]) for i in range(self.nb_rows())])

  def merge(self, r) -> 'DataFrame':
    if not isinstance(r, DataFrame):
      raise TypeError("Can not merge {} with {}.".format(__class__, type(r)))
    return DataFrame(pd.concat([self.d, r.d], axis=0))



import numpy as np
from typing import Iterable, Tuple


def crop(data: np.ndarray,
         crop_sizes: Iterable[Tuple[int, int]]) -> np.ndarray:
    """
    """
    def expand_ints(crop_sizes):
      result = []
      for cs = crop_sizes:
        if cs is None:
          result.append((0, 0))
        elif isinstance(cs, int):
          result.append((cs, cs))
        else:
          result.append(cs)
      return result



from .data import Data
import dask


class DataFrame(Data):
    pass


class DataIterableDask(Data):
    def map(self, func) -> 'DataIterableDask':
        return DataIterableDask(dask.delayed(map)(func, self.data()))


class DataIterableWithUnknownKeys(Data):
    def select(self, key):
        return DataFrame(self.d.get_group(key))

    def drop_keys(self) -> ResultsDask:
        return self.map(lambda x: x[1])



import numpy as np
import pandas as pd
from ._ssim import ssim
from ._msssim import msssim
from ._basic import psnr, rmse


def _metrics_name(m):
  return {
      psnr: 'psnr',
      rmse: 'rmse',
      ssim: 'ssim',
      msssim: 'msssim',
      np.mean: 'mean',
      np.max: 'max',
      np.min: 'min',
      np.std: 'std'
  }[m]


def crop(value, vmin=0.0, vmax=None):
  if vmin is not None:
    value = np.minimum(value, vmin)
  if vmax is not None:
    value = np.maximum(value, vmax)
  return value


def rescale(target, label=None, k=None, bias=None, vkeep=None):
  if vkeep is None:
    vkeep = 0.0
  if label is not None:
    k = np.mean(label - vkeep) / np.mean(target - vkeep)
    if abs(k) > 1e-5:
      bias = (vkeep - k * vkeep) / k
  return k * (target + bias)


def compare_analysis(label,
                     target,
                     metrics=tuple(),
                     *,
                     max_val=None,
                     backend=None):
  if label.ndim != 4:
    raise ValueError("Label should has shape [N,H,W,C], got {}.".format(
        label.shape))
  if target.ndim != 4:
    raise ValueError("Target should has shape [N,H,W,C], got {}.".format(
        target.shape))
  if label.shape != target.shape:
    raise ValueError(
        "Label and target should have same shape, got label: {}, target: {}.".
        format(label.shape, target.shape))
  if max_val is None:
    max_val = np.max(label)
  results = {}
  for m in metrics:
    name = _metrics_name(m)
    if m in (psnr, rmse):
      result = m(label, target)
    elif m in (ssim, msssim):
      result = m(label, target, max_val=max_val, backend=backend)
    else:
      raise ValueError("Unknown metric {}.".format(m))
    results[name] = np.array(result)
  return pd.DataFrame(data=results)


def single_analysis(target, metrics=(np.mean, np.min, np.max, np.std)):
  if target.ndim != 4:
    raise ValueError("Target should has shape [N,H,W,C], got {}.".format(
        target.shape))
  nb_images = target.shape[0]
  results = {}
  for m in metrics:
    name = _metrics_name(m)
    if m in (np.mean, np.max, np.min, np.std):
      results[name] = m(target, axis=(1, 2, 3))
  return pd.DataFrame(data=results)



from ..core import Data, DataIterable
from ..pandas_ import Series, DataFrame
from ..geometry import Vector3
from enum import Enum
import pandas as pd
import numpy as np


class ParticleID(Enum):
    Gamma = 22


class ColumnNames:
    Event = 'eventID'
    Process = 'processName'
    Particle = 'particalID'
    SourceX = 'srcX'
    SourceY = 'srcY'
    SourceZ = 'srcZ'
    X = 'x'
    Y = 'y'
    Z = 'z'
    EnergyDeposit = 'energy'


class RootStep(Series):
    def position(self) -> Vector3:
        return Vector3(self.columns(ColumnNames.X,
                                    ColumnNames.Y,
                                    ColumnNames.Z))

    def source_position(self) -> Vector3:
        return Vector3(self.columns(ColumnNames.SourceX,
                                    ColumnNames.SourceY,
                                    ColumnNames.SourceZ))

    def energy_deposit(self) -> Data:
        return self.columns(ColumnNames.EnergyDeposit).first()


class RootData(DataFrame):
    def split_by_event(self) -> DataIterable:
        return (self.split_by(ColumnNames.Event)
                .drop_keys()
                .map(lambda d: d.to(Event)))

    def to_event(self) -> 'Event':
        if not ColumnNames.Event in self.d.columns:
            return Event(self.d)

    def first(self):
        return super().first().to(RootStep)


class EnergyDeposit(Data):
    def __init__(self, position, energy):
        super().__init__((position, energy))

    def position(self):
        return self.d[0]

    def energy(self):
        return self.d[1]


class Event(DataFrame):
    def __init__(self, dataframe: pd.DataFrame):
        super().__init__(dataframe)

    def first_step(self):
        return self.first().to(RootStep)

    def split_by_step(self) -> DataIterable:
        return self.split_row().map(lambda d: d.to(RootStep))

    def source_position(self) -> Vector3:
        return self.first_step().source_position()

    def first_position(self) -> Vector3:
        return self.first_step().position()

    def incident_direction(self) -> Vector3:
        dp = self.first_position().d - self.source_position().d
        return Vector3(tuple(dp / np.linalg.norm(dp)))

    def energy_deposit_list(self) -> DataIterable:
        return self.split_by_step().map(lambda s: EnergyDeposit(s.position(),
                                                             s.energy_deposit()))



import numpy as np


class Image:
  """
  Unified Image object, with data shape [H, W, 3] or [H, W, 1].
  """

  @classmethod
  def boxing(cls, data):
    pass

  @classmethod
  def unboxing(cls, data):
    pass

  def __init__(self, data):
    self._data = data


class StackedImages:
  """
  with data shape [N, H, W, C]
  """

  def __init__(self, data):
    """
    :type data: numpy.ndarray or dxl.data.Image 
    """
    self._data = data

  def _load_data(self):
    if isinstance(data, Image):
      self._data = data.reshape([1] + list(data.shape))
    if isinstance(data, (list, tuple)):
      if isinstance(data[0], Image):
        self._data = np.array()

  def ndim(self):
    return self._data.ndim


class DictOfStackedImages:
  def __init__(self, data):
    self._data = self._unified_data(data)

  def _unified_data(self, data):
    if isinstance(data, dict):
      result = {}
      for k, v in data.items():
        if v.ndim == 3:
          result[k] = np.expand_dims(v, 3)
        else:
          result[k] = v
      return result
    else:
      return data

  def data(self):
    return self._data

  def shape(self):
    return {k: self.data()[k].shape for k in self.data()}

  def ndim(self):
    return {k: self.data()[k].ndim for k in self.data()}


class ImagesGrid:
  """
  with data shape [R, C][H, W, C], for visualization use only.
  """
  pass


from dxl.fs import Path
from .core import Data
import pandas as pd


class DataCSV(Data):
  def __init__(self, path: str):
    super().__init__(Path(path))

  def load(self, cls=None) -> 'dxl.data.DataFrame':
    from .pandas_ import DataFrame
    if cls is None:
      cls = DataFrame
    if cls == DataFrame:
      return DataFrame(pd.read_csv(self.d.s))



def slices_from_ranges(ranges):
    pass


def _parse_single_str_slice(s):
    if ':' in s:
        splitted = s.split(':')
        splitted = [None if s == '' else int(s) for s in splitted]
        return slice(*splitted)
    else:
        return int(s)


def slices_from_str(slices_str: str):
    slices = slices_str[1:-1].split(',')
    return tuple(map(_parse_single_str_slice, slices))


__all__ = ['slices_from_ranges', 'slices_from_str']


import numpy as np


def fspecial(size, sigma):
  """Function to mimic the 'fspecial' gaussian MATLAB function.
  :returns result: 2d ndarray, gaussian filter. with shape `[1, size, size, 1]`.
  :type result: Numpy array
  """
  radius = size // 2
  offset = 0.0
  start, stop = -radius, radius + 1
  if size % 2 == 0:
    offset = 0.5
    stop -= 1
  x, y = np.mgrid[offset + start:stop, offset + start:stop]
  assert len(x) == size
  g = np.exp(-((x**2 + y**2) / (2.0 * sigma**2)))
  result = g / g.sum()
  result = result.reshape([result.shape[0], result.shape[1], 1, 1])
  return result


import numpy as np
from typing import Tuple


class Sinogram2D:
  def __init__(self, data: np.ndarray):
    """
    Sinogram2D processings.

    :param data: Raw sinogram, 2 dimensional array, in the order [0..nb_views, 0..nb_sensors]
    :type data: NumPy array
    """
    self._data = np.array(data)

  @property
  def d(self):
    return self._data

  def crop(self, x_crops: Tuple[int, int],
           y_crops: Tuple[int, int] = None) -> 'Sinogram2D':
    if isinstance(x_crops, int):
      x_crops = (x_crops, x_crops)
    if isinstance(y_crops, int):
      y_crops = (y_crops, y_crops)
    if x_crops[1] > 0 and y_crops[1] > 0:
      return Sinogram2D(self.d[x_crops[0]:-x_crops[1], y_crops[0]:-y_crops[1]])
    if x_crops[1] > 0:
      return Sinogram2D(self.d[x_crops[0]:-x_crops[1], y_crops[0]:])
    if y_crops[1] > 0:
      return Sinogram2D(self.d[x_crops[0]:, y_crops[0]:-y_crops[1]])
    return Sinogram2D(self.d[x_crops[0]:, y_crops[0]:])


from ..core import Data, DataNDArray, DataIterable
import numpy as np


class VectorLowDim(Data):

    dim = None

    def __init__(self, data):
        """
        Parameters:

        - `data`: VectorLowDim, list/tuple or numpy.ndarray of correspongding dimension.

        Raises:

        - `TypeError` if given data with unsupported type.
        - `ValueError` if given data with invalid dimension.
        """
        if isinstance(data, Data):
            data = data.d
        if isinstance(data, (list, tuple)):
            data = np.array(data)
        if not isinstance(data, np.ndarray):
            raise TypeError("Unsupported datatype {}.".format(type(data)))
        if data.size != np.max(data.shape):
            raise ValueError("Invalid data shape: {}.".format(data.shape))
        if self.dim is not None and data.size != self.dim:
            fmt = "Invalid data dimension {} when {} is expected for {}."
            raise ValueError(fmt.format(data.size, self.dim, __class__))
        super().__init__(np.array(data))


class Vector1(VectorLowDim):
    dim = 1

    def x(self):
        return Data(self.d[0])


class Vector2(VectorLowDim):
    dim = 2

    def x(self):
        return Data(self.d[0])

    def y(self):
        return Data(self.d[1])


class Vector3(VectorLowDim):
    dim = 3

    def x(self):
        return Data(self.d[0])

    def y(self):
        return Data(self.d[1])

    def z(self):
        return Data(self.d[2])


class GridSpec:
    def __init__(self, start, end, step=None):
        self.start = np.array(start)
        self.end = np.array(end)
        if step is not None:
            self.step = np.array(step)
        else:
            self.step = None

    def spec_of_dim(self, dim: int):
        if self.step is not None:
            return self.start[dim], self.end[dim], self.step[dim]
        else:
            return self.start[dim], self.end[dim]

    def dim(self):
        return self.start.size

    def shape(self):
        result = []
        for i in self.dim():
            s, e, h = self.spec_of_dim(i)
            result.append(int((e - s) / h))
        return result


def grid_index(samples, gird_spec: GridSpec):
    samples = np.array(samples)
    result = np.zeros(samples.shape, np.int32)
    for idim in range(samples.shape[1]):
        start, end, step = gird_spec.spec_of_dim(idim)
        result[:, idim] = (np.floor((samples[:, idim] - start) / step)
                           .astype(np.int32))
    return result


def inbound_mask(samples, grid_spec: GridSpec):
    samples = np.array(samples)
    result = np.empty([samples.shape[0]], bool)
    for i in range(samples.shape[0]):
        result[i] = True
        for j in range(samples.shape[1]):
            start, end, _ = grid_spec.spec_of_dim(j)
            if samples[i, j] < start or samples[i, j] >= end:
                result[i] = False
                break
    return result


class Points(DataNDArray):
    """
    Data of VectorLowDim.
    """

    def __init__(self, data):
        super().__init__(data)
        if self.dim() > 3 or self.dim() < 1:
            raise ValueError("Invalid data shape {}.".format(self.d.shape))

    def nb_samples(self):
        return self.d.shape[0]

    def dim(self):
        return self.d.shape[1]

    def to_list(self) -> DataIterable:
        cls = [None, Vector1, Vector2, Vector3][self.dim()]
        return DataIterable([cls(self.d[i, ...]) for i in range(self.nb_samples())])

    def grid_spec(self, start: VectorLowDim, end: VectorLowDim, step: VectorLowDim) -> DataNDArray:
        start = VectorLowDim.from_(start)
        end = VectorLowDim.from_(end)
        step = VectorLowDim.from_(step)
        return GridSpec(start.d, end.d, step.d)

    def histogram_index(self, grid_spec: GridSpec) -> DataNDArray:
        result = grid_index(self.d, gird_spec)
        mask = inbound_mask(self.d, grid_spec)
        return result[mask, :]

    def histogram(self, grid_spec: GridSpec) -> DataNDArray:
        index = self.histogram_index(grid_spec)
        result = np.zeros(grid_spec.shape())
        for i in range(index.shape[0]):
            result[list(index[i, :])] += 1
        return DataNDArray(result)


class PointsWithValue(DataNDArray):
    """
    Points with an extra dims as function dims.
    Thus self.d[:, :pdim] are data for VectorLowDim, and self.d[:, pdim:] are 
    data for function values.
    """

    def __init__(self, point, value=None, dim_point=None):
        """
        Parameters:

        - `point`: DataNDArray or numpy.ndarray, ndarry for points only or 
        combination of point and value. In the latter case, the leading
        `dim_point` columns are treated as point dims.

        - `value`: DataNDArray or numpy.ndarray, if `None`, value will be
        gathered from point.

        - `dim_point`: dimension of point, will be ignored if `value` is provided.
        If `None`, use `point.shape[1] - 1`.
        """
        data = self.convert_data_to_ndarray(point)
        if value is not None:
            value = self.convert_data_to_ndarray(value)
            dim_point = data.shape[1]
            data = np.concatenate([data, value], axis=1)
        super().__init__(data)
        if dim_point is None:
            dim_point = data.shape[1] - 1
            self._squeeze = True
        else:
            self._squeeze = False
        self._dim_point = dim_point

    def dim_point(self):
        return self._point_dim

    def dim_value(self):
        return self.data().shape[1] - self.dim_point()

    def nb_samples(self):
        return self.data().shape[0]

    def histogram(self, grid_spec: GridSpec) -> DataNDArray:
        index = grid_index(self.d[:, :self.dim_point()], gird_spec)
        mask = inbound_mask(self.d[:, :self.dim_point()], grid_spec)
        new_shape = list(grid_spec.shape()) + [1] * self.dim_value()
        result = np.zeros(new_shape)
        for i in range(index.shape[0]):
            if not mask[i]:
                continue
            result[list(index[i, :])] += self.d[i, self.dim_point():]
        return DataNDArray(result)



import numpy as np


def rmse_(label, target):
  err = target - label
  value = np.sqrt(np.sum(np.square(err)) / label.size)
  # base = np.sqrt(np.sum(np.square(target)) / label.size)
  # return value / base
  return value


def mean_absolute_error_(label, target):
  err = target - label
  return np.mean(np.abs(err))


def bias_(label, target):
  return np.mean(label - target)


def variance_(label, target):
  return np.std(label - target)


def psnr_(label, target):
  minv = np.min(label)
  maxv = np.max(label)
  sca = 255.0 / (maxv - minv)
  ln = (label - minv) * sca
  tn = (target - minv) * sca
  rmv = rmse(ln, tn)
  value = 10 * np.log((255.0**2) / (rmv**2)) / np.log(10)
  return value


def _map_batch(func, label, target):
  nb_images = label.shape[0]
  result = [
      func(label[i, :, :, 0], target[i, :, :, 0]) for i in range(nb_images)
  ]
  return np.array(result)


def _unified_dim_2_and_4(func, label, target):
  if label.ndim == 2:
    return func(label, target)
  elif label.ndim == 4:
    return _map_batch(func, label, target)
  raise ValueError("Invalid shape: {}.".format(label.shape))


def psnr(label, target):
  return _unified_dim_2_and_4(psnr_, label, target)


def rmse(label, target):
  return _unified_dim_2_and_4(rmse_, label, target)


def mean_absolute_error(label, target):
  return _unified_dim_2_and_4(mean_absolute_error_, label, target)


def variance(label, target):
  return _unified_dim_2_and_4(variance_, label, target)


def bias(label, target):
  return _unified_dim_2_and_4(bias_, label, target)


from ..utils.slices import slices_from_str
from ._numpy import load_npz
from ._hdf5 import load_h5
import numpy as np
from collections import UserDict


class NDArraySpec(UserDict):
    class KEYS:
        PATH_FILE = 'path_file'
        PATH_DATASET = 'path_dataset'
        SLICES = 'slices'

    @property
    def path_file(self):
        return self.data.get(self.KEYS.PATH_FILE)

    @property
    def path_dataset(self):
        return self.data.get(self.KEYS.PATH_DATASET)

    @property
    def slices(self):
        return self.data.get(self.KEYS.SLICES)


def load_array(spec):
    path_file = spec.get('path_file')
    path_dataset = spec.get('path_dataset')
    slices = spec.get('slices')

    def maybe_slice(a):
        if slices is None:
            return a
        else:
            return a[slices_from_str(slices)]

    if path_file is None:
        raise TypeError("path_file not found in spec: {}.".format(spec))
    if path_file.endswith('.npy'):
        return maybe_slice(np.load(path_file))
    if path_file.endswith('.npz'):
        return maybe_slice(load_npz(path_file)[path_dataset])
    if path_file.endswith('.h5'):
        return load_h5(path_file, path_dataset, slices)



from .gate import ParticleID, ColumnNames, RootData
from ..text import DataCSV
from ..core import DataNDArray
from dxl.fs import File
import pandas as pd
import numpy as np


def gamma_incident_directions(csv_filename) -> pd.DataFrame:
    events = (DataCSV('hits.csv').load()
              .to(RootData)
              .split_by_event())
    inci_dirs = (events
                 .map_call('incident_direction')
                 .map(lambda d: d.d)
                 .to(DataNDArray))
    return pd.DataFrame(data=np.array(inci_dirs.d), columns=['x', 'y', 'z'])


def gamma_energy_deposit_samples(csv_filename):
    events = (DataCSV('hits.csv').load()
              .to(RootData)
              .split_by_event())
    eps = events.map_call('energy_deposit_list').flatten()
    poss = eps.map_call('position')
    engs = eps.map_call('energy')
    result = (poss.zip(engs)
              .map(lambda t: tuple(list(t[0].data()) + [t[1]]))
              .to(DataNDArray))
    return pd.DataFrame(data=np.array(result.d),
                        columns=['x', 'y', 'z', 'energy'])



import numpy as np
import tensorflow as tf


def _ssim_cpu(img1, img2, window, c1, c2):
  from scipy import signal
  from scipy.ndimage.filters import convolve
  from scipy.signal import convolve2d
  window = np.transpose(window, [2, 0, 1, 3])
  mu1 = signal.fftconvolve(img1, window, mode='valid')
  mu2 = signal.fftconvolve(img2, window, mode='valid')
  sigma11 = signal.fftconvolve(img1 * img1, window, mode='valid')
  sigma22 = signal.fftconvolve(img2 * img2, window, mode='valid')
  sigma12 = signal.fftconvolve(img1 * img2, window, mode='valid')
  return _ssim_cpu_post(mu1, mu2, sigma11, sigma22, sigma12, c1, c2)


def _ssim_empty(img1, img2, c1, c2):
  mu1, mu2 = img1, img2
  sigma11 = img1 * img1
  sigma22 = img2 * img2
  sigma12 = img1 * img2
  return _ssim_cpu_post(mu1, mu2, sigma11, sigma22, sigma12, c1, c2)


def _ssim_cpu_post(mu1, mu2, sigma11, sigma22, sigma12, c1, c2):
  mu11 = mu1 * mu1
  mu22 = mu2 * mu2
  mu12 = mu1 * mu2
  sigma11 -= mu11
  sigma22 -= mu22
  sigma12 -= mu12
  v1 = 2.0 * sigma12 + c2
  v2 = sigma11 + sigma22 + c2
  ssim = np.mean(
      (((2.0 * mu12 + c1) * v1) / ((mu11 + mu22 + c1) * v2)), axis=(1, 2, 3))
  cs = np.mean(v1 / v2, axis=(1, 2, 3))
  return ssim, cs


def _ssim_tensorflow(img1, img2, window, c1, c2):
  from dxl.learn.core import Session
  mu1 = tf.nn.conv2d(img1, window, [1] * 4, 'VALID')
  mu2 = tf.nn.conv2d(img2, window, [1] * 4, 'VALID')
  sigma11 = tf.nn.conv2d(img1 * img1, window, [1] * 4, 'VALID')
  sigma22 = tf.nn.conv2d(img2 * img2, window, [1] * 4, 'VALID')
  sigma12 = tf.nn.conv2d(img1 * img2, window, [1] * 4, 'VALID')
  mu11 = mu1 * mu1
  mu22 = mu2 * mu2
  mu12 = mu1 * mu2
  sigma11 -= mu11
  sigma22 -= mu22
  sigma12 -= mu12

  # Calculate intermediate values used by both ssim and cs_map.

  v1 = 2.0 * sigma12 + c2
  v2 = sigma11 + sigma22 + c2
  ssim = tf.reduce_mean(
      (((2.0 * mu12 + c1) * v1) / ((mu11 + mu22 + c1) * v2)), axis=[1, 2, 3])
  cs = tf.reduce_mean(v1 / v2, axis=[1, 2, 3])
  if isinstance(img1, np.ndarray):
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    with tf.Session(config=config) as sess:
      ssim = sess.run(ssim)
      cs = sess.run(cs)
  return ssim, cs


def ssim(img1,
         img2,
         max_val,
         filter_size=11,
         filter_sigma=1.5,
         k1=0.01,
         k2=0.03,
         backend='cpu'):
  """Return the Structural Similarity Map between `img1` and `img2`.
  This function attempts to match the functionality of ssim_index_new.m by
  Zhou Wang: http://www.cns.nyu.edu/~lcv/ssim/msssim.zip
  Arguments:
    img1: Numpy array holding the first RGB image batch.
    img2: Numpy array holding the second RGB image batch.
    max_val: the dynamic range of the images (i.e., the difference between the
      maximum the and minimum allowed values).
    filter_size: Size of blur kernel to use (will be reduced for small images).
    filter_sigma: Standard deviation for Gaussian blur kernel (will be reduced
      for small images).
    k1: Constant used to maintain stability in the SSIM calculation (0.01 in
      the original paper).
    k2: Constant used to maintain stability in the SSIM calculation (0.03 in
      the original paper).
  Returns:
    Pair containing the mean SSIM and contrast sensitivity between `img1` and
    `img2`.
  Raises:
    RuntimeError: If input images don't have the same shape or don't have four
      dimensions: [batch_size, height, width, depth].
  """
  from ..utils import fspecial
  if isinstance(img1, np.ndarray):
    if img1.shape != img2.shape:
      raise RuntimeError('Input images must have the same shape (%s vs. %s).',
                         img1.shape, img2.shape)
    if img1.ndim != 4:
      raise RuntimeError('Input images must have four dimensions, not %d',
                         img1.ndim)
    img1 = img1.astype(np.float32)
    img2 = img2.astype(np.float32)
    _, height, width, _ = img1.shape
  elif isinstance(img1, tf.Tensor):
    _, height, width, _ = img1.shape.as_list()
  size = min(filter_size, height, width)

  # Scale down sigma if a smaller filter size is used.
  sigma = size * filter_sigma / filter_size if filter_size else 0
  c1 = (k1 * max_val)**2
  c2 = (k2 * max_val)**2
  if filter_size:
    window = fspecial(size, sigma)
    if backend == 'cpu':
      return _ssim_cpu(img1, img2, window, c1, c2)
    elif backend == 'tensorflow':
      return _ssim_tensorflow(img1, img2, window, c1, c2)
    raise ValueError("Unknown backend {}.".format(backend))
  else:
    return _ssim_empty(img1, img2, window, c1, c2)



import numpy as np
from scipy import signal
import tensorflow as tf

from ..utils import fspecial
from ._ssim import ssim


def _msssim_common(img1, img2, max_val, filter_size, filter_sigma, k1, k2,
                   weights, backend):
  levels = weights.size
  mssim = []
  mcs = []
  if backend == 'cpu':
    g = _msssim_cpu
  else:
    g = _msssim_tensorflow
  for im1, im2 in g(img1, img2, levels):
    _ssim, _cs = ssim(
        im1,
        im2,
        max_val=max_val,
        filter_size=filter_size,
        filter_sigma=filter_sigma,
        k1=k1,
        k2=k2,
        backend=backend)
    mssim.append(_ssim)
    mcs.append(_cs)
  result = 1.0
  for mc, w in zip(mcs[:-1], weights[:-1]):
    result = result * mc**w
  result = result * mssim[-1]**weights[-1]
  if isinstance(result, tf.Tensor):
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    with tf.Session(config=config) as sess:
      result = sess.run(result)
  return result


def _msssim_cpu(img1, img2, levels):
  from scipy.ndimage.filters import convolve
  downsample_filter = np.ones((1, 2, 2, 1)) / 4.0
  im1, im2 = [x.astype(np.float64) for x in [img1, img2]]
  for _ in range(levels):
    yield im1, im2
    filtered = [
        convolve(im, downsample_filter, mode='reflect') for im in [im1, im2]
    ]
    im1, im2 = [x[:, ::2, ::2, :] for x in filtered]


def _msssim_tensorflow(img1, img2, levels):
  downsample_filter = np.ones((2, 2, 1, 1)) / 4.0
  im1, im2 = [tf.constant(x, dtype=tf.float32) for x in [img1, img2]]
  for _ in range(levels):
    yield im1, im2
    filtered = [
        tf.nn.conv2d(im, downsample_filter, [1, 2, 2, 1], 'SAME')
        for im in [im1, im2]
    ]
    im1, im2 = filtered


def msssim(img1,
           img2,
           max_val=255,
           filter_size=11,
           filter_sigma=1.5,
           k1=0.01,
           k2=0.03,
           weights=None,
           backend=None):
  """Return the MS-SSIM score between `img1` and `img2`.
  This function implements Multi-Scale Structural Similarity (MS-SSIM) Image
  Quality Assessment according to Zhou Wang's paper, "Multi-scale structural
  similarity for image quality assessment" (2003).
  Link: https://ece.uwaterloo.ca/~z70wang/publications/msssim.pdf
  Author's MATLAB implementation:
  http://www.cns.nyu.edu/~lcv/ssim/msssim.zip
  Arguments:
    img1: Numpy array holding the first RGB image batch.
    img2: Numpy array holding the second RGB image batch.
    max_val: the dynamic range of the images (i.e., the difference between the
      maximum the and minimum allowed values).
    filter_size: Size of blur kernel to use (will be reduced for small images).
    filter_sigma: Standard deviation for Gaussian blur kernel (will be reduced
      for small images)
    k1: Constant used to maintain stability in the SSIM calculation (0.01 in
      the original paper).
    k2: Constant used to maintain stability in the SSIM calculation (0.03 in
      the original paper).
    weights: List of weights for each level; if none, use five levels and the
      weights from the original paper.
  Returns:
    MS-SSIM score between `img1` and `img2`.
  Raises:
    RuntimeError: If input images don't have the same shape or don't have four
      dimensions: [batch_size, height, width, depth].
  """
  if backend is None:
    backend = 'cpu'
  if img1.shape != img2.shape:
    raise RuntimeError('Input images must have the same shape (%s vs. %s).',
                       img1.shape, img2.shape)
  if img1.ndim != 4:
    raise RuntimeError('Input images must have four dimensions, not %d',
                       img1.ndim)

  # Note: default weights don't sum to 1.0 but do match the paper / matlab code.
  weights = np.array(weights
                     if weights else [0.0448, 0.2856, 0.3001, 0.2363, 0.1333])
  return _msssim_common(img1, img2, max_val, filter_size, filter_sigma, k1, k2,
                        weights, backend)



import h5py
import numpy as np
import pandas as pd


def _load_h5_data(path_file, path_dataset):
  with h5py.File(path_file, 'r') as fin:
    data = np.array(fin[path_dataset])
    if data.ndim == 3:
      data = np.expand_dims(data, 3)
    return data


def _to_ndarray(input_):
  if input_ is None:
    return input_
  if isinstance(input_, np.ndarray):
    return input_
  if isinstance(input_, dict):
    if input_['type'] == 'h5':
      return _load_h5_data(input_['path_file'], input_['path_dataset'])
  raise ValueError("Can't convert {} to ndrray.".format(input_))


def metric(name_or_method):
  from .image import psnr, ssim, msssim, rmse
  if isinstance(name_or_method, str):
    return {
        'mean': np.mean,
        'max': np.max,
        'min': np.min,
        'std': np.std,
        'ssim': ssim,
        'psnr': psnr,
        'msssim': msssim,
        'rmse': rmse,
    }[name_or_method]
  return name_or_method


def analysis(target,
             label=None,
             single_analysis_metrics=None,
             compare_analysis_metrics=None,
             output_csv=None,
             *,
             denorm=None):
  from .image import single_analysis, compare_analysis
  target = _to_ndarray(target)
  if denorm is not None:
    target = target * denorm['std'] + denorm['mean']
  label = _to_ndarray(label)
  if single_analysis_metrics is not None:
    single_analysis_metrics = [metric(s) for s in single_analysis_metrics]
  if compare_analysis_metrics is not None:
    compare_analysis_metrics = [metric(m) for m in compare_analysis_metrics]
  results = []
  if single_analysis_metrics is not None:
    results.append(single_analysis(target, single_analysis_metrics))
  if compare_analysis_metrics is not None:
    results.append(compare_analysis(label, target, compare_analysis_metrics))
  result = pd.concat(results, axis=1)
  if output_csv is not None:
    result.to_csv(output_csv)
  return result


def analysis_with_config(config_file):
  import json
  with open(config_file, 'r') as fin:
    c = json.load(fin)
  target = c.get('target')
  label = c.get('label')
  single_analysis_metrics = c.get('single')
  compare_analysis_metrics = c.get('compare')
  if single_analysis_metrics is not None:
    single_analysis_metrics = [metric(m) for m in single_analysis_metrics]
  if compare_analysis_metrics is not None:
    compare_analysis_metrics = [metric(m) for m in compare_analysis_metrics]
  return analysis(target, label, single_analysis_metrics,
                  compare_analysis_metrics, output_csv)



from ._base import PersistentData, PersistentDataInDataset
from abc import ABCMeta, abstractmethod, abstractproperty
import h5py
import numpy as np


class NDArray:
    def __init__(self, data, slices=None):
        self._data = data
        self._slices = slices

    def data(self):
        """
        Unboxing data.
        """
        return self._data

    def to_numpy_ndarray(self):
        d = self.data()
        if isinstance(d, np.ndarray):
            return d
        else:
            return np.array(d)

    @property
    def shape(self):
        return self._data.shape

    @property
    def ndim(self):
        return self._data.ndim

    def __getitem__(self, val):
        return self._data[val]


class NDArrayHDF5(PersistentDataInDataset, NDArray):
    """
    Provide an abstract layer of dataset/ndarray object in HDF5 file.
    Currently support:

    - `self.shape`
    - `self.ndim`
    - `self[slices]`
    
    Load data lazily, thus only after `self.data()` is called,
    the whole array is loaded into memory.
    """

    def __init__(self, path_file, path_dataset, slices=None):
        """
        """
        PersistentDataInDataset.__init__(self, path_file, path_dataset)
        # Lazy loading data. Thus NDArray is initilized with `data` = None
        NDArray.__init__(self, None, slices)

    def data(self):
        if self._data is None:
            from dxl.data.io import load_h5
            self._data = load_h5(self.path, self.path_dataset, self._slices)
        return self._data

    def _apply_to_h5dataset_if_data_is_none(self, func):
        if self._data is None:
            with h5py.File(self.path, 'r') as fin:
                return func(fin[self.path_dataset])
        else:
            return func(fin[self._data])

    @property
    def shape(self):
        return self._apply_to_h5dataset_if_data_is_none(lambda d: d.shape)

    @property
    def ndim(self):
        return self._apply_to_h5dataset_if_data_is_none(lambda d: d.ndim)

    def __getitem__(self, val):
        return self._apply_to_h5dataset_if_data_is_none(lambda d: d[val])


__all__ = ['NDArrayHDF5']


from setuptools import setup, find_packages
setup(
    name='dxl-data',
    version='0.0.13',
    description='Data Processing Library.',
    url='https://github.com/Hong-Xiang/dxdata',
    author='Hong Xiang',
    author_email='hx.hongxiang@gmail.com',
    license='MIT',
    namespace_packages=['dxl'],
    packages=find_packages('src/python'),
    package_dir={'': 'src/python'},
    install_requires=['dxl-core', 'dxl-fs', 'pandas==0.20.3'],
    scripts=[],
    entry_points="""
        [console_scripts]
        dxdata=dxl.data.cli.main:dxdata
    """,
    zip_safe=False)



import numpy as np


def __guess_shape(length, ndim=None):
  if length == 1 and (ndim is None or ndim == 0):
    return []
  if ndim is None:
    return [length]
  size = int(np.ceil(np.power(length, 1 / ndim)))
  if size**ndim == length:
    return [size] * ndim
  raise ValueError("Cannot guess shape of length {} with ndim {}.".format(
      length, ndim))


def load_bin(filename, shape=None, dtype=np.float32, *, fs=None, ndim=None):
  open_func = open if fs is None else fs.open
  with open_func(filename, 'rb') as fin:
    raw_data = np.fromstring(fin.read(), dtype)
  if shape is None:
    shape = __guess_shape(raw_data.size, ndim)
  return np.reshape(raw_data, shape)


def save_bin(filename, tensor, dtype=np.float32):
  tensor = tensor.astype(dtype)
  raw_data = tensor.tostring()
  with open(filename, 'wb') as fout:
    fout.write(raw_data)



from typing import Dict
import numpy as np
import h5py


def h5_add_dataset(h5group, dct_of_ndarray, batch_dim=0, tqdm=None):
    """
    Add dict of numpy ndarray to hdf5 group/file.

    :param batch_dim: dimension which is expanded when calculating chunks,
                        e.g. for a ndarray with shape `[10, 5, 5, 1]`, and `batch_dim = 0`,
                        the chunks will be `[1, 5, 5, 1]`.
    """
    if tqdm is None:
        to_iter = dct_of_ndarray
    else:
        to_iter = tqdm(dct_of_ndarray, leave=False)
    for k in to_iter:
        print('Saving', k)
        if isinstance(dct_of_ndarray[k], dict):
            g = h5group.create_group(k)
            h5_add_dataset(g, dct_of_ndarray[k], batch_dim, tqdm)
        else:
            if batch_dim == 0:
                chunk_size = [1] + list(dct_of_ndarray[k].shape[1:])
                nb_chunks = dct_of_ndarray[k].shape[0] // chunk_size[0]
                # Fix for h5py memory leak
                if nb_chunks > 10000:
                    chunk_size[0] = dct_of_ndarray[k].shape[0] // 10000
                chunk_size = tuple(chunk_size)
            else:
                raise ValueError("Batchdim != 0 is not implemented yet.")
            h5group.create_dataset(
                k,
                data=dct_of_ndarray[k],
                shape=dct_of_ndarray[k].shape,
                dtype=dct_of_ndarray[k].dtype,
                chunks=tuple(chunk_size),
                compression="gzip")


def save_h5(file_path,
            dataset: Dict[str, np.ndarray],
            dataset_path=None,
            *,
            tqdm=None):
    """
    Save dict of ndarray to hdf5 file.

    - `file_path` path of hdf5 file in filesystem.
    - `dataset` dict of np.ndarray to be saved.
    - `dataset_path` in file group path, if is `None`, `'/'` will be used.
    - `tqdm` progress bar to be used
    """
    dataset_path = dataset_path or '/'
    with h5py.File(file_path) as fout:
        g = fout.require_group(dataset_path)
        h5_add_dataset(g, dataset, tqdm=tqdm)


def load_h5(path_file, path_dataset=None, slices=None):
    """
    Load numpy.ndarray from HDF5 file.
    
    Args:
    - `path_file`: `str` or `pathlib.Path`, path of hdf5 file
    - `path_dataset`: dataset path in file 
    - `slices`: tuple of slice objects or `None`, or str
    """
    if isinstance(slices, str):
        from ..utils.slices import slices_from_str
        slices = slices_from_str(slices)
    if path_dataset is None:
        path_dataset = '/'
    with h5py.File(path_file, 'r') as fin:
        if slices is None:
            return np.array(fin[path_dataset])
        else:
            return np.array(fin[path_dataset][slices])


class PersistentData:
    def __init__(self, path):
        self._path = path

    @property
    def path(self):
        return self._path


class PersistentDataInDataset(PersistentData):
    def __init__(self, path_file, path_dataset):
        super().__init__(path_file)
        self._path_dataset = path_dataset

    @property
    def path_dataset(self):
        return self._path_dataset



"""
=====================
Data Analysis Library
=====================

Introduction
============

Module: dxl.data

Data analysis libraries is designed to provide easy to use and reuseable tools
for data analysis, i.e. "data cleaning".

An object oriented warp on data is provided, and valid operations to these data
were implemented as member method. It can be called in syntax like ::

    data = (CSVFile('data.csv')
            .load()
            .map_to(Person)
            .split_by('name')
            .select('Tim'))

By composing reuseable blocks, we are trying to achieve process of "data cleaning"
with following properties (with decreasing priority):

1.  Easy to develop, thus provide useful API and provide more hints by member 
methods.

2.  Reuseability. Try to achieve DRY as much as possible.

3.  Scalable. Able to be running on multiple computers, thus aiming to used in
cloud or through micro service.

4.  Fast.

"""

from .core import *